from django.templatetags.static import static
from django.utils.html import format_html

from wagtail.core import hooks


@hooks.register('insert_editor_js')
def editor_js():
    html = '<script type="text/javascript" src="{}"></script>'.format(
        static('wagtailautocomplete/dist.js')
    )
    return format_html(html)


@hooks.register('insert_editor_css')
def editor_css():
    html = '<link rel="stylesheet" type="text/css" href="{}" />'.format(
        static('wagtailautocomplete/dist.css')
    )
    return format_html(html)



SECRET_KEY = 'NOTSECRET'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    }
}

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'wagtail.core',
    'wagtail.documents',
    'wagtail.images',
    'wagtail.admin',
    'wagtail.sites',
    'wagtail.users',
    'taggit',
    'wagtailautocomplete',
    'wagtailautocomplete.tests.testapp',
)

MIDDLEWARE = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.request',
            ],
        },
    },
]

ROOT_URLCONF = 'wagtailautocomplete.tests.testapp.urls'

STATIC_URL = '/static/'
MEDIA_URL = '/media/'



import warnings

from django.apps import apps
from django.db.models import ManyToManyField
from wagtail.admin.edit_handlers import FieldPanel

from .widgets import Autocomplete


def _can_create(model):
    """Returns True if the given model has implemented the autocomplete_create
    method to allow new instances to be creates from a single string value.
    """
    return callable(getattr(
        model,
        'autocomplete_create',
        None,
    ))


class AutocompletePanel(FieldPanel):
    def __init__(self, field_name, target_model=None, **kwargs):
        # For compatability with old 'page_type' argument
        if 'page_type' in kwargs:
            warnings.warn(
                'page_type argument has been replaced with target_model',
                DeprecationWarning
            )
            target_model = kwargs.pop('page_type', None)
        if 'is_single' in kwargs:
            warnings.warn('is_single argument is no longer used', DeprecationWarning)
        self.target_model_kwarg = target_model

        kwargs.pop('is_single', None)  # Deprecated kwarg
        super().__init__(field_name, **kwargs)

    def clone(self):
        return self.__class__(
            field_name=self.field_name,
            target_model=self.target_model_kwarg,
        )

    @property
    def is_single(self):
        # Should cover all many-to-many relationships
        return not issubclass(self.model._meta.get_field(self.field_name).__class__,
                              ManyToManyField)

    @property
    def target_model(self):
        if not self.target_model_kwarg:
            return self.model._meta.get_field(self.field_name).remote_field.model
        elif isinstance(self.target_model_kwarg, str):
            return apps.get_model(self.target_model_kwarg)
        return self.target_model_kwarg

    def on_model_bound(self):
        target_model = self.target_model
        can_create = _can_create(target_model)
        self.widget = type(
            '_Autocomplete',
            (Autocomplete,),
            dict(target_model=target_model, can_create=can_create, is_single=self.is_single),
        )



from django.conf.urls import url
try:
    from wagtail.admin.decorators import require_admin_access
except ImportError:
    from wagtail.admin.auth import require_admin_access

from wagtailautocomplete.views import create, objects, search

urlpatterns = [
    url(r'^create/', require_admin_access(create)),
    url(r'^objects/', require_admin_access(objects)),
    url(r'^search/', require_admin_access(search)),
]



from django.conf.urls import include, url

from wagtail.admin import urls as wagtailadmin_urls
from wagtail.core import urls as wagtail_urls

from wagtailautocomplete.urls.admin import urlpatterns as autocomplete_admin_urls
from wagtailautocomplete.urls.public import urlpatterns as autocomplete_public_urls


urlpatterns = [
    url(r'^admin/autocomplete/', include(autocomplete_admin_urls)),
    url(r'^admin/', include(wagtailadmin_urls)),
    url(r'^autocomplete/', include(autocomplete_public_urls)),
    url(r'', include(wagtail_urls)),
]



from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.test import TestCase
from wagtail.core.models import Site

from wagtailautocomplete.tests.testapp.models.page_to_page import (
    MultipleAutocompletePage, SingleAutocompletePage, TargetPage)


User = get_user_model()


class ObjectsViewTestCase(TestCase):
    def test_autocomplete_missing_pks_param(self):
        """The objects view should return a Bad Request response if not
        given the pks parameter.

        """
        response = self.client.get("/autocomplete/objects/")
        assert response.status_code == 400

    def test_target_model_not_found(self):
        """The objects view should return a Bad Request response if not
        given a valid model type.

        """
        response = self.client.get(
            "/autocomplete/objects/", {"pks": "1,2", "type": "<invalid type>"}
        )
        assert response.status_code == 400

    def test_invalid_pks(self):
        """The objects view should return a Bad Request response if not
        given the valid primary keys.

        """
        invalid = "abcde"
        response = self.client.get("/autocomplete/objects/", {"pks": invalid})
        assert response.status_code == 400


class SearchViewTestCase(TestCase):
    def setUp(self):
        self.site = Site.objects.get(is_default_site=True)
        self.root_page = self.site.root_page
        self.target_page1 = TargetPage(title="target1 cars")
        self.target_page2 = TargetPage(title="target2 cars")
        self.root_page.add_child(instance=self.target_page1)
        self.root_page.add_child(instance=self.target_page2)
        self.single_page = SingleAutocompletePage(
            title="Autocomplete singly.",
            target=self.target_page1,
        )
        self.multi_page = MultipleAutocompletePage(
            title="Autocomplete multiply.",
        )
        self.root_page.add_child(instance=self.single_page)
        self.root_page.add_child(instance=self.multi_page)
        self.multi_page.targets.add(self.target_page1, self.target_page2)

    def test_target_model_not_found(self):
        """The search view should return a Bad Request response if not
        given a valid model type.

        """
        response = self.client.get(
            "/autocomplete/search/", {"type": "<invalid type>"})
        self.assertEqual(response.status_code, 400)

    def test_invalid_limit(self):
        """The search view should return Bad Request if not given
        a numeric query limit.

        """
        invalid = "abcde"
        response = self.client.get("/autocomplete/search/", {"limit": invalid})
        self.assertEqual(response.status_code, 400)

    def test_search_blank_single_exception_ignored(self):
        """The search view should handle a blank exclude clause."""
        response = self.client.get(
            "/autocomplete/search/"
            "?type=testapp.TargetPage"
            "&query=cars"
            "&exclude="
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()['items']), 2)

    def test_search_blank_multi_exceptions_ignored(self):
        """The search view should handle multiple blank exclude clauses."""
        response = self.client.get(
            "/autocomplete/search/"
            "?type=testapp.TargetPage"
            "&query=cars"
            "&exclude=,,,"
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()['items']), 2)

    def test_search_valid_exception(self):
        response = self.client.get(
            "/autocomplete/search/"
            "?type=testapp.TargetPage"
            "&query=cars"
            "&exclude={},102,103".format(self.target_page1.pk)
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.json()['items']), 1)
        self.assertEqual(response.json()['items'][0]['title'], 'target2 cars')


class CreateViewTestCase(TestCase):
    def setUp(self):
        self.superuser = User.objects.create(
            username="testuser", password="unusable", is_superuser=True
        )
        self.adminuser = User.objects.create(
            username="testuser2", password="unusable")
        # Permission to log into the admin, but not to change
        # anything.
        admin_permission = Permission.objects.get(codename="access_admin")
        self.adminuser.user_permissions.add(admin_permission)

    def test_missing_value_param(self):
        """The create view should return a Bad Request response if not
        given the value parameter.

        """
        self.client.force_login(self.superuser)
        response = self.client.post("/admin/autocomplete/create/")
        assert response.status_code == 400

    def test_target_model_not_found(self):
        """The create view should return a Bad Request response if not
        given a valid model type.

        """
        self.client.force_login(self.superuser)
        response = self.client.post(
            "/admin/autocomplete/create/",
            {"value": "a", "type": "<invalid type>"}
        )
        assert response.status_code == 400

    def test_user_lacks_permissions(self):
        """The create view should return a Forbidden response if the user does
        not have create permission for that model.

        """
        self.client.force_login(self.adminuser)
        response = self.client.post(
            "/admin/autocomplete/create/", {"value": "a"})
        assert response.status_code == 403

    def test_autocomplete_create_not_implemented(self):
        """The create view should return a Bad Request response if
        `autocomplete_create` is not implemented on the model requested.

        """
        self.client.force_login(self.superuser)
        response = self.client.post(
            "/admin/autocomplete/create/", {"value": "a"})
        assert response.status_code == 400



from urllib.parse import unquote

from django.apps import apps
from django.contrib.contenttypes.models import ContentType
from django.http import (HttpResponseBadRequest, HttpResponseForbidden,
                         JsonResponse)
from django.views.decorators.http import require_GET, require_POST


def render_page(page):
    if getattr(page, 'specific', None):
        # For support of non-Page models like Snippets.
        page = page.specific
    if callable(getattr(page, 'autocomplete_label', None)):
        title = page.autocomplete_label()
    else:
        title = page.title
    return dict(pk=page.pk, title=title)


@require_GET
def objects(request):
    pks_param = request.GET.get('pks')
    if not pks_param:
        return HttpResponseBadRequest()
    target_model = request.GET.get('type', 'wagtailcore.Page')
    try:
        model = apps.get_model(target_model)
    except Exception:
        return HttpResponseBadRequest()

    try:
        pks = [
            unquote(pk)
            for pk in pks_param.split(',')
        ]
        queryset = model.objects.filter(pk__in=pks)
    except Exception:
        return HttpResponseBadRequest()

    if getattr(queryset, 'live', None):
        # Non-Page models like Snippets won't have a live/published status
        # and thus should not be filtered with a call to `live`.
        queryset = queryset.live()

    results = map(render_page, queryset)
    return JsonResponse(dict(items=list(results)))


@require_GET
def search(request):
    search_query = request.GET.get('query', '')
    target_model = request.GET.get('type', 'wagtailcore.Page')
    try:
        model = apps.get_model(target_model)
    except Exception:
        return HttpResponseBadRequest()

    try:
        limit = int(request.GET.get('limit', 100))
    except ValueError:
        return HttpResponseBadRequest()

    field_name = getattr(model, 'autocomplete_search_field', 'title')
    filter_kwargs = dict()
    filter_kwargs[field_name + '__icontains'] = search_query
    queryset = model.objects.filter(**filter_kwargs)

    if getattr(queryset, 'live', None):
        # Non-Page models like Snippets won't have a live/published status
        # and thus should not be filtered with a call to `live`.
        queryset = queryset.live()

    exclude = request.GET.get('exclude', '')
    if exclude:
        exclusions = [unquote(item) for item in exclude.split(',') if item]
        queryset = queryset.exclude(pk__in=exclusions)

    results = map(render_page, queryset[:limit])
    return JsonResponse(dict(items=list(results)))


@require_POST
def create(request, *args, **kwargs):
    value = request.POST.get('value', None)
    if not value:
        return HttpResponseBadRequest()

    target_model = request.POST.get('type', 'wagtailcore.Page')
    try:
        model = apps.get_model(target_model)
    except Exception:
        return HttpResponseBadRequest()

    content_type = ContentType.objects.get_for_model(model)
    permission_label = '{}.add_{}'.format(
        content_type.app_label,
        content_type.model
    )
    if not request.user.has_perm(permission_label):
        return HttpResponseForbidden()

    method = getattr(model, 'autocomplete_create', None)
    if not callable(method):
        return HttpResponseBadRequest()

    instance = method(value)
    return JsonResponse(render_page(instance))



import json

from django.forms import Widget
from wagtail.admin.edit_handlers import widget_with_script

from .views import render_page


class Autocomplete(Widget):
    template_name = 'wagtailautocomplete/autocomplete.html'

    def get_context(self, *args, **kwargs):
        context = super(Autocomplete, self).get_context(*args, **kwargs)
        context['widget']['target_model'] = self.target_model._meta.label
        context['widget']['can_create'] = self.can_create
        context['widget']['is_single'] = self.is_single
        return context

    def format_value(self, value):
        if not value:
            return 'null'
        if type(value) == list:
            return json.dumps([
                render_page(page)
                for page in self.target_model.objects.filter(pk__in=value)
            ])
        else:
            return json.dumps(render_page(self.target_model.objects.get(pk=value)))

    def value_from_datadict(self, data, files, name):
        value = json.loads(data.get(name))
        if not value:
            return None

        if type(value) == list:
            return [obj['pk'] for obj in value]

        return value['pk']

    def render(self, name, value, attrs=None, renderer=None):
        return widget_with_script(
            super(Autocomplete, self).render(name, value, attrs, renderer),
            self.render_js_init(attrs['id']))

    def render_js_init(self, id):
        return "window.initAutoCompleteWidget('{0}');".format(id)



from setuptools import setup, find_packages
from os import path
from wagtailautocomplete import __version__

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst')) as f:
    long_description = f.read()


setup(
    name='wagtail-autocomplete',
    version=__version__,

    packages=find_packages(),
    include_package_data=True,

    description='An Autocomplete edit handler for Pages, Snippets, and more.',
    long_description=long_description,

    url='https://github.com/wagtail/wagtail-autocomplete',

    author='Emily Horsman and Harris Lapiroff',
    author_email='me@emilyhorsman.com',

    license='BSD-3-Clause',

    install_requires=[
        'wagtail>=2.3',
    ],

    extras_require={
        'docs': [
            'Sphinx>=1.7',
            'sphinx_rtd_theme>=0.4.0',
        ],
        'test': [
            'tox',
            'pytest>=3.5',
            'pytest-django>=3.2',
            'beautifulsoup4>=4.8',
            'html5lib>=0.999999999',
            'pytest-pythonpath>=0.7.2',
        ],
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django',
        'Framework :: Wagtail',
        'Framework :: Wagtail :: 2',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



from bs4 import BeautifulSoup

from django.test import TestCase
from django.contrib.auth import get_user_model

from wagtail.core.models import Site

from .testapp.models.page_to_page import SingleAutocompletePage


User = get_user_model()


class PageToPageTestCase(TestCase):
    def setUp(self):
        self.site = Site.objects.get()
        self.root_page = self.site.root_page
        self.superuser = User.objects.create(
            username='testuser',
            password='unusable',
            is_superuser=True,
        )

    def test_autocomplete_widget_renders(self):
        """
        There should be exactly one autocomplete input on the create page for
        a new SingleAutocompletePage
        """

        self.client.force_login(self.superuser)
        create_page_url = '/admin/pages/add/{app_label}/{model_name}/{root_pk}/'.format(
            app_label=SingleAutocompletePage._meta.app_label,
            model_name=SingleAutocompletePage._meta.model_name,
            root_pk=self.site.root_page.pk,
        )
        response = self.client.get(create_page_url)
        # Drill down through the html looking for a properly rendered
        # Autocomplete Panel
        soup = BeautifulSoup(response.content, 'html5lib')
        autocomplete_inputs = soup.select('[data-autocomplete-input]')
        self.assertEqual(len(autocomplete_inputs), 1)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.13 on 2018-06-19 15:39
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import modelcluster.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wagtailcore', '0040_page_draft_title'),
    ]

    operations = [
        migrations.CreateModel(
            name='MultipleAutocompletePage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='MultipleOptionalAutocompletePage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='SingleAutocompletePage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='SingleOptionalAutocompletePage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.CreateModel(
            name='TargetPage',
            fields=[
                ('page_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='wagtailcore.Page')),
            ],
            options={
                'abstract': False,
            },
            bases=('wagtailcore.page',),
        ),
        migrations.AddField(
            model_name='singleoptionalautocompletepage',
            name='target',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='testapp.TargetPage'),
        ),
        migrations.AddField(
            model_name='singleautocompletepage',
            name='target',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='testapp.TargetPage'),
        ),
        migrations.AddField(
            model_name='multipleoptionalautocompletepage',
            name='targets',
            field=modelcluster.fields.ParentalManyToManyField(blank=True, to='testapp.TargetPage'),
        ),
        migrations.AddField(
            model_name='multipleautocompletepage',
            name='targets',
            field=modelcluster.fields.ParentalManyToManyField(to='testapp.TargetPage'),
        ),
    ]



from django.db import models
from modelcluster.fields import ParentalManyToManyField

from wagtailautocomplete.edit_handlers import AutocompletePanel

try:
    # Wagtail 2.x
    from wagtail.core.models import Page
except ImportError:
    # Wagtail 1.x
    from wagtail.wagtailcore.models import Page


class TargetPage(Page):
    pass


class SingleAutocompletePage(Page):
    target = models.ForeignKey(
        TargetPage, on_delete=models.PROTECT
    )

    content_panels = Page.content_panels + [
        AutocompletePanel('target', target_model='testapp.TargetPage'),
    ]


class SingleOptionalAutocompletePage(Page):
    target = models.ForeignKey(
        TargetPage, blank=True, null=True, on_delete=models.PROTECT
    )


class MultipleAutocompletePage(Page):
    targets = ParentalManyToManyField(TargetPage)


class MultipleOptionalAutocompletePage(Page):
    targets = ParentalManyToManyField(TargetPage, blank=True)
